/* $Id: ContentHandlerToXMLStreamWriter.java,v 1.1 2004/11/03 01:18:59 dandiep Exp $
*
* Copyright (c) 2004, Sun Microsystems, Inc.
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are
* met:
*
*     * Redistributions of source code must retain the above copyright
*      notice, this list of conditions and the following disclaimer.
*
*     * Redistributions in binary form must reproduce the above
*      copyright notice, this list of conditions and the following
*       disclaimer in the documentation and/or other materials provided
*       with the distribution.
*
*     * Neither the name of Sun Microsystems, Inc. nor the names of its
*       contributors may be used to endorse or promote products derived
*       from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
* "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
* LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
* A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
* OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
* DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
* THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
* OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.codehaus.xfire.xmlbeans.util;

import org.xml.sax.Attributes;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamWriter;
import java.util.Stack;

/**
* This is a simple utility class that adapts SAX events into StAX
* {@link javax.xml.stream.XMLStreamWriter} events, bridging between
* the two parser technologies.
*
* This ContentHandler does not own the XMLStreamWriter.  Therefore, it will
* not close or flush the writer at any point.
* 
* @author Ryan.Shoemaker@Sun.COM
* @version 1.0
*/
public class ContentHandlerToXMLStreamWriter extends DefaultHandler {

   // SAX events will be sent to this XMLStreamWriter
   private final XMLStreamWriter staxWriter;

   // storage for prefix bindings
   private final Stack prefixBindings;

   private boolean writeStart;
   
   public ContentHandlerToXMLStreamWriter(XMLStreamWriter staxCore, boolean writeStart) {
       this.staxWriter = staxCore;
       prefixBindings = new Stack(); // default of 10 seems reasonable
       this.writeStart = writeStart;
   }

   /*
    * (non-Javadoc)
    * 
    * @see org.xml.sax.ContentHandler#endDocument()
    */
   public void endDocument() throws SAXException {
       try {

           staxWriter.flush();
       } catch (XMLStreamException e) {
           throw new SAXException(e);
       }
   }

   /*
    * (non-Javadoc)
    * 
    * @see org.xml.sax.ContentHandler#startDocument()
    */
   public void startDocument() throws SAXException {

   }

   /*
    * (non-Javadoc)
    * 
    * @see org.xml.sax.ContentHandler#characters(char[], int, int)
    */
   public void characters(char[] ch, int start, int length)
       throws SAXException {

       try {
           staxWriter.writeCharacters(ch, start, length);
       } catch (XMLStreamException e) {
           throw new SAXException(e);
       }

   }

   /*
    * (non-Javadoc)
    * 
    * @see org.xml.sax.ContentHandler#ignorableWhitespace(char[], int, int)
    */
   public void ignorableWhitespace(char[] ch, int start, int length)
       throws SAXException {
       
       characters(ch,start,length);
   }

   /*
    * (non-Javadoc)
    * 
    * @see org.xml.sax.ContentHandler#endPrefixMapping(java.lang.String)
    */
   public void endPrefixMapping(String prefix) throws SAXException 
   {
       // TODO: no-op?

       // I think we can ignore these SAX events because StAX
       // automatically scopes the prefix bindings.
   }

   /*
    * (non-Javadoc)
    * 
    * @see org.xml.sax.ContentHandler#skippedEntity(java.lang.String)
    */
   public void skippedEntity(String name) throws SAXException {
       try {
           staxWriter.writeEntityRef(name);
       } catch (XMLStreamException e) {
           throw new SAXException(e);
       }
   }

   /*
    * (non-Javadoc)
    * 
    * @see org.xml.sax.ContentHandler#setDocumentLocator(org.xml.sax.Locator)
    */
   public void setDocumentLocator(Locator locator) {
       // TODO: no-op?
       // there doesn't seem to be any way to pass location info
       // along to the XMLStreamWriter. On the XMLEventWriter side, you
       // can set the location info on the event objects.
   }

   /*
    * (non-Javadoc)
    * 
    * @see org.xml.sax.ContentHandler#processingInstruction(java.lang.String,
    *      java.lang.String)
    */
   public void processingInstruction(String target, String data)
       throws SAXException {

       try {
           staxWriter.writeProcessingInstruction(target, data);
       } catch (XMLStreamException e) {
           throw new SAXException(e);
       }

   }

   /*
    * (non-Javadoc)
    * 
    * @see org.xml.sax.ContentHandler#startPrefixMapping(java.lang.String,
    *      java.lang.String)
    */
   public void startPrefixMapping(String prefix, String uri)
       throws SAXException {

       if (prefix.equals("xml")) {
           return;
       }

       // defend against parsers that pass null in for "xmlns" prefix
       if (prefix == null) {
           prefix = "";
       }

       prefixBindings.add(prefix);
       prefixBindings.add(uri);
   }

   /*
    * (non-Javadoc)
    * 
    * @see org.xml.sax.ContentHandler#endElement(java.lang.String,
    *      java.lang.String, java.lang.String)
    */
   public void endElement(String namespaceURI, String localName, String qName)
       throws SAXException {

       try {
           // TODO: is this all we have to do?
           staxWriter.writeEndElement();
       } catch (XMLStreamException e) {
           throw new SAXException(e);
       }
   }

    public void startElement(String uri, String localName,
            String qName, Attributes atts) throws SAXException
    {
        //System.out.println(qName + " " + uri);
        try
        {
            String prefix = getPrefix(qName);
            //System.out.println(prefix);
            
            if ( uri != null )
            {
                if ( prefix.equals("") )
                {
                    String defNS = staxWriter.getNamespaceContext().getNamespaceURI("");
                    if ( !defNS.equals(uri) )
                    {
                        staxWriter.setDefaultNamespace(uri);
                        staxWriter.writeStartElement( uri, localName );
                        staxWriter.writeDefaultNamespace( uri );
                    }
                    else
                    {
                        staxWriter.writeStartElement( uri, localName );
                    }  
                }
                else
                {
                    staxWriter.setPrefix( prefix, uri );
                    
                    staxWriter.writeStartElement( 
                        prefix,
                        localName,
                        uri);
                }
            }
            else
            {
                staxWriter.writeStartElement( localName );
            }
            
            String nsuri, nsprefix;
            while (prefixBindings.size() != 0)
            {
                nsuri = (String) prefixBindings.pop();
                nsprefix = (String) prefixBindings.pop();
                if (nsprefix.length() != 0)
                {
                    //System.out.println("writing " + nsprefix + " : " + nsuri);
                    // this method handles "", null, and "xmlns" prefixes
                    // properly
                    staxWriter.writeNamespace(nsprefix, nsuri);
                }
            }
            staxWriter.flush();
            //writeAttributes(atts);
        }
        catch (XMLStreamException e)
        {
            throw new RuntimeException(e);
        }

    }

    /**
     * Generate a StAX writeAttribute event for each attribute
     * 
     * @param atts
     *            attributes from the SAX event
     */
    private void writeAttributes(Attributes atts) throws XMLStreamException
    {
        for (int i = 0; i < atts.getLength(); i++)
        {
            System.out.println("writing attribute " + atts.getQName(i));
            staxWriter.writeAttribute(getPrefix(atts.getQName(i)), atts
                    .getURI(i), atts.getLocalName(i), atts.getValue(i));
        }
    }

    /**
     * Pull the prefix off of the specified QName.
     * 
     * @param qName
     *            the QName
     * @return the prefix or the empty string if it doesn't exist.
     */
    private String getPrefix(String qName)
    {
        int idx = qName.indexOf(':');
        if (idx == -1)
        {
            return "";
        }
        else
        {
            return qName.substring(0, idx);
        }
    }

}
